/*  XMMS - Cross-platform multimedia player
 *  Copyright (C) 1998-2002  Peter Alm, Mikael Alm, Olle Hallnas,
 *                           Thomas Nilsson and 4Front Technologies
 *  Copyright (C) 1999-2004  Haavard Kvaalen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdarg.h>
#include <string.h>
#include <errno.h>
#include <libxmms/xentry.h>
#include <gdk/gdkkeysyms.h>
#include "mpg123.h"

#define MAX_STR_LEN 100
#define MAX_ENTRY_LEN2 1023

static GtkWidget *window = NULL;
static GtkWidget *notebook = NULL;
static GtkWidget *filename_entry, *id3v1_frame, *id3v2_frame;
static GtkWidget *v1_checkbox, *v2_checkbox;
static GtkWidget *v1_title_entry, *v1_artist_entry, *v1_album_entry, *v1_year_entry, *v1_tracknum_entry, *v1_comment_entry;
static GtkWidget *v2_title_entry, *v2_artist_entry, *v2_album_entry, *v2_year_entry, *v2_tracknum_entry, *v2_comment_entry,
				 *v2_composer_entry, *v2_orig_artist_entry, *v2_url_entry, *v2_encoded_by_entry;
static GtkWidget *v1_genre_combo, *v2_genre_combo;
static GtkWidget *mpeg_level, *mpeg_bitrate, *mpeg_samplerate, *mpeg_flags;
static GtkWidget *mpeg_fileinfo;

static GPtrArray *v1_labels_list = NULL, *v2_labels_list = NULL;    // TODO: Where will these be freed?
static GList *genre_list;
struct genre_item {
	const char *name;
	int id;
};
static int v1_current_genre;
static int v2_current_genre;
static char *current_filename;

extern char *mpg123_filename;
extern int mpg123_bitrate, mpg123_frequency, mpg123_layer, mpg123_lsf, mpg123_mode;
extern gboolean mpg123_stereo, mpg123_mpeg25;

static void label_set_text(GtkWidget * label, const char *str, ...)
G_GNUC_PRINTF(2, 3);

static void set_entry_tag_v1(GtkEntry * entry, const char * tag, int length)
{
	char *text = g_strchomp(g_strndup(tag, length));
	gtk_entry_set_text(entry, text);
	g_free(text);
}

static void get_entry_tag_v1(GtkEntry * entry, char * tag, int length)
{
	strncpy(tag, gtk_entry_get_text(entry), length);
}

void copy_entry_tag_v1(GtkEntry * src, GtkEntry * dest, int length)
{
	set_entry_tag_v1(dest, gtk_entry_get_text(src), length);
	return;
}

static void set_entry_tag_v2(GtkEntry * entry, struct id3_tag* id3, guint32 frame_type)
{
	struct id3_frame* frame;

	frame = id3_get_frame(id3, frame_type, 1);
	if (frame != NULL)
	{
		char *text, *url, *comment;

		text = id3_get_text(frame);
		if (text != NULL)
		{
			gtk_entry_set_text(entry, text);
			g_free(text);
			return;
		}

		url = id3_get_url(frame);
		if (url != NULL)
		{
			gtk_entry_set_text(entry, url);
			g_free(url);
			return;
		}

		comment = id3_get_comment(frame);
		if (comment != NULL)
		{
			gtk_entry_set_text(entry, comment);
			g_free(url);
			return;
		}
	}
}

void copy_entry_tag_v2(GtkEntry * src, GtkEntry * dest)
{
	gtk_entry_set_text(dest, gtk_entry_get_text(src));
	return;
}

static int genre_find_index(GList *genre_list, int id)
{
	int idx = 0;
	while (genre_list)
	{
		struct genre_item *item = genre_list->data;
		if (item->id == id)
			break;
		idx++;
		genre_list = genre_list->next;
	}
	if (!genre_list)
		return 0;
	return idx;
}

static int genre_find_index_str(GList *genre_list, const char* str)
{
	int idx = 0;
	while (genre_list)
	{
		struct genre_item *item = genre_list->data;
		if (strcmp(item->name, str) == 0)
			break;
		idx++;
		genre_list = genre_list->next;
	}
	if (!genre_list)
		return 0;
	return idx;
}

static int genre_comp_func(gconstpointer a, gconstpointer b)
{
	const struct genre_item *ga = a, *gb = b;
	return strcasecmp(ga->name, gb->name);
}

static void remove_id3v1(void)
{
	int fd, len;
	struct id3v1tag_t tag;
	char *msg = NULL;

	if (!strncasecmp(current_filename, "http://", 7))
		return;

	if ((fd = open(current_filename, O_RDWR)) != -1)
	{
		len = lseek(fd, -128, SEEK_END);
		read(fd, &tag, sizeof (struct id3v1tag_t));

		if (!strncmp(tag.tag, "TAG", 3))
		{
			if (ftruncate(fd, len))
				msg = g_strdup_printf(_("%s\n"
							"Unable to truncate file: %s"),
						      _("Couldn't remove tag!"),
						      strerror(errno));
		}
		close(fd);
	}
	else
		msg = g_strdup_printf(_("%s\nUnable to open file: %s"),
				      _("Couldn't remove tag!"),
				      strerror(errno));
	if (msg)
	{
		GtkWidget *mwin = xmms_show_message(_("File Info"), msg, _("OK"),
						    FALSE, NULL, NULL);
		gtk_window_set_transient_for(GTK_WINDOW(mwin),
					     GTK_WINDOW(window));
		g_free(msg);
	}
}

static void save_cb(GtkWidget * w, gpointer data)
{
	int fd;
	struct id3v1tag_t tag;
	char *msg = NULL;

	if (!strncasecmp(current_filename, "http://", 7))
		return;

	if ((fd = open(current_filename, O_RDWR)) != -1)
	{
		if (!GTK_TOGGLE_BUTTON(v1_checkbox)->active) {
			/* Try to save id3v1 tag */
			int tracknum;

			lseek(fd, -128, SEEK_END);
			read(fd, &tag, sizeof (struct id3v1tag_t));

			if (!strncmp(tag.tag, "TAG", 3))
				lseek(fd, -128, SEEK_END);
			else
				lseek(fd, 0, SEEK_END);
			tag.tag[0] = 'T';
			tag.tag[1] = 'A';
			tag.tag[2] = 'G';
			get_entry_tag_v1(GTK_ENTRY(v1_title_entry), tag.title, 30);
			get_entry_tag_v1(GTK_ENTRY(v1_artist_entry), tag.artist, 30);
			get_entry_tag_v1(GTK_ENTRY(v1_album_entry), tag.album, 30);
			get_entry_tag_v1(GTK_ENTRY(v1_year_entry), tag.year, 4);
			tracknum = atoi(gtk_entry_get_text(GTK_ENTRY(v1_tracknum_entry)));
			if (tracknum > 0)
			{
				get_entry_tag_v1(GTK_ENTRY(v1_comment_entry), tag.u.v1_1.comment, 28);
				tag.u.v1_1.__zero = 0;
				tag.u.v1_1.track_number = MIN(tracknum, 255);
			}
			else
				get_entry_tag_v1(GTK_ENTRY(v1_comment_entry),
                        tag.u.v1_0.comment, 30);
			tag.genre = v1_current_genre;
			if (write(fd, &tag, sizeof (tag)) != sizeof (tag))
				msg = g_strdup_printf(_("%s\nUnable to write to file: %s"),
				                      _("Couldn't write tag!"),
					              strerror(errno));
		} else {
			/* Remove the id3v1 tag from the file */
			remove_id3v1();
		}

		if (!GTK_TOGGLE_BUTTON(v2_checkbox)->active) {
		        struct id3_tag* id3;

			lseek(fd, SEEK_SET, 0);
		        id3 = id3_open_fd(fd, 0);
		        if (id3 == NULL)
		           id3 = id3_new();

		        if (id3 != NULL)
		        {
				id3_set_text(id3_get_or_add_frame(id3, ID3_TIT2), gtk_entry_get_text(GTK_ENTRY(v2_title_entry)));
				id3_set_text(id3_get_or_add_frame(id3, ID3_TPE1), gtk_entry_get_text(GTK_ENTRY(v2_artist_entry)));
				id3_set_text(id3_get_or_add_frame(id3, ID3_TALB), gtk_entry_get_text(GTK_ENTRY(v2_album_entry)));
				id3_set_text(id3_get_or_add_frame(id3, ID3_TYER), gtk_entry_get_text(GTK_ENTRY(v2_year_entry)));
				id3_set_comment(id3_get_or_add_frame(id3, ID3_COMM), gtk_entry_get_text(GTK_ENTRY(v2_comment_entry)));
				id3_set_text(id3_get_or_add_frame(id3, ID3_TRCK), gtk_entry_get_text(GTK_ENTRY(v2_tracknum_entry)));
				if (v2_current_genre != 0xff)
				{
					/* Essentially the same behavior as Winamp2's ID3v2 tagger */
					char genre[255];
					snprintf(genre, sizeof(genre), "(%d)%s",
					         v2_current_genre,
					         gtk_entry_get_text(GTK_ENTRY(GTK_COMBO(v2_genre_combo)->entry)));
					id3_set_text(id3_get_or_add_frame(id3, ID3_TCON), genre);
				}
				else
				{
					id3_set_text(id3_get_or_add_frame(id3, ID3_TCON), gtk_entry_get_text(GTK_ENTRY(GTK_COMBO(v2_genre_combo)->entry)));
				}
				id3_set_text(id3_get_or_add_frame(id3, ID3_TCOM), gtk_entry_get_text(GTK_ENTRY(v2_composer_entry)));
				id3_set_url(id3_get_or_add_frame(id3, ID3_WCOM), gtk_entry_get_text(GTK_ENTRY(v2_url_entry)));
				id3_set_text(id3_get_or_add_frame(id3, ID3_TOPE), gtk_entry_get_text(GTK_ENTRY(v2_orig_artist_entry)));
				id3_set_text(id3_get_or_add_frame(id3, ID3_TENC), gtk_entry_get_text(GTK_ENTRY(v2_encoded_by_entry)));

				id3_write_tag_filename(id3, current_filename);
			}

		} else {
			id3_remove_tag_filename(current_filename);
		}

		if (fd)
			close(fd);
	}
	else
		msg = g_strdup_printf(_("%s\nUnable to open file: %s"),
				      _("Couldn't write tag!"),
				      strerror(errno));
	if (msg)
	{
		GtkWidget *mwin = xmms_show_message(_("File Info"), msg, _("OK"),
						    FALSE, NULL, NULL);
		gtk_window_set_transient_for(GTK_WINDOW(mwin),
					     GTK_WINDOW(window));
		g_free(msg);
	}
}

static void label_set_text(GtkWidget * label, const char *str, ...)
{
	va_list args;
	char tempstr[MAX_STR_LEN];

	va_start(args, str);
	g_vsnprintf(tempstr, MAX_STR_LEN, str, args);
	va_end(args);

	gtk_label_set_text(GTK_LABEL(label), tempstr);
}

static void set_mpeg_level_label(gboolean mpeg25, int lsf, int layer)
{
	if (mpeg25)
		label_set_text(mpeg_level, "MPEG 2.5, layer %d", layer);
	else
		label_set_text(mpeg_level, "MPEG %d, layer %d", lsf + 1, layer);
}

static char* channel_mode_name(int mode)
{
	static const char *channel_mode[] =
		{N_("Stereo"), N_("Joint stereo"),
		 N_("Dual channel"), N_("Single channel")};
	if (mode < 0 || mode > 3)
		return "";
	return gettext(channel_mode[mode]);
}

static void file_info_http(char *filename)
{
	gtk_widget_set_sensitive(id3v1_frame, FALSE);
	gtk_widget_set_sensitive(id3v2_frame, FALSE);
	if (mpg123_filename && !strcmp(filename, mpg123_filename) &&
	    mpg123_bitrate != 0)
	{
		set_mpeg_level_label(mpg123_mpeg25, mpg123_lsf,
                mpg123_layer);
		label_set_text(mpeg_bitrate, _("Bitrate: %d kb/s"),
                mpg123_bitrate);
		label_set_text(mpeg_samplerate, _("Samplerate: %d Hz"),
                mpg123_frequency);
		label_set_text(mpeg_flags, "%s",
                channel_mode_name(mpg123_mode));
	}
}

void copy_v2_to_v1_cb(GtkButton *button, gpointer user_data)
{
    copy_entry_tag_v1(GTK_ENTRY(v2_title_entry),   GTK_ENTRY(v1_title_entry), 30);
    copy_entry_tag_v1(GTK_ENTRY(v2_artist_entry),  GTK_ENTRY(v1_artist_entry), 30);
    copy_entry_tag_v1(GTK_ENTRY(v2_album_entry),   GTK_ENTRY(v1_album_entry), 30);
    copy_entry_tag_v1(GTK_ENTRY(v2_year_entry),    GTK_ENTRY(v1_year_entry), 4);
    copy_entry_tag_v1(GTK_ENTRY(v2_comment_entry), GTK_ENTRY(v1_comment_entry), 30);
    copy_entry_tag_v1(GTK_ENTRY(v2_tracknum_entry), GTK_ENTRY(v1_tracknum_entry), 3);

    gtk_list_select_item(GTK_LIST(GTK_COMBO(v1_genre_combo)->list), genre_find_index(genre_list, v2_current_genre));
    return;
}

void copy_v1_to_v2_cb(GtkButton *button, gpointer user_data)
{
    copy_entry_tag_v2(GTK_ENTRY(v1_title_entry),   GTK_ENTRY(v2_title_entry));
    copy_entry_tag_v2(GTK_ENTRY(v1_artist_entry),  GTK_ENTRY(v2_artist_entry));
    copy_entry_tag_v2(GTK_ENTRY(v1_album_entry),   GTK_ENTRY(v2_album_entry));
    copy_entry_tag_v2(GTK_ENTRY(v1_year_entry),    GTK_ENTRY(v2_year_entry));
    copy_entry_tag_v2(GTK_ENTRY(v1_comment_entry), GTK_ENTRY(v2_comment_entry));
    copy_entry_tag_v2(GTK_ENTRY(v1_tracknum_entry), GTK_ENTRY(v2_tracknum_entry));

    gtk_list_select_item(GTK_LIST(GTK_COMBO(v2_genre_combo)->list), genre_find_index(genre_list, v1_current_genre));
    return;
}

void v1_toggle_cb (GtkWidget *widget, gpointer data)
{
    int i = 0;
    if (GTK_TOGGLE_BUTTON (widget)->active)
    {
        // If control reaches here, the toggle button is down
        // Gray out labels
        for (i = 0; i < v1_labels_list->len; i++) {
            gtk_widget_set_sensitive(GTK_WIDGET( g_ptr_array_index(v1_labels_list, i) ), FALSE);
        }
        gtk_widget_set_sensitive(v1_title_entry, FALSE);
        gtk_widget_set_sensitive(v1_artist_entry, FALSE);
        gtk_widget_set_sensitive(v1_album_entry, FALSE);
        gtk_widget_set_sensitive(v1_year_entry, FALSE);
        gtk_widget_set_sensitive(v1_tracknum_entry, FALSE);
        gtk_widget_set_sensitive(v1_comment_entry, FALSE);
        gtk_widget_set_sensitive(v1_genre_combo, FALSE);
    } else {

        // If control reaches here, the toggle button is up
        // Enable labels
        for (i = 0; i < v1_labels_list->len; i++) {
            gtk_widget_set_sensitive(GTK_WIDGET( g_ptr_array_index(v1_labels_list, i) ), TRUE);
        }
        gtk_widget_set_sensitive(v1_title_entry, TRUE);
        gtk_widget_set_sensitive(v1_artist_entry, TRUE);
        gtk_widget_set_sensitive(v1_album_entry, TRUE);
        gtk_widget_set_sensitive(v1_year_entry, TRUE);
        gtk_widget_set_sensitive(v1_tracknum_entry, TRUE);
        gtk_widget_set_sensitive(v1_comment_entry, TRUE);
        gtk_widget_set_sensitive(v1_genre_combo, TRUE);
    }
}

void v2_toggle_cb (GtkWidget *widget, gpointer data)
{
    int i = 0;
    if (GTK_TOGGLE_BUTTON (widget)->active)
    {
        // If control reaches here, the toggle button is down
        // Gray out labels
        for (i = 0; i < v2_labels_list->len; i++) {
            gtk_widget_set_sensitive(GTK_WIDGET( g_ptr_array_index(v2_labels_list, i) ), FALSE);
        }
        gtk_widget_set_sensitive(v2_title_entry, FALSE);
        gtk_widget_set_sensitive(v2_artist_entry, FALSE);
        gtk_widget_set_sensitive(v2_album_entry, FALSE);
        gtk_widget_set_sensitive(v2_year_entry, FALSE);
        gtk_widget_set_sensitive(v2_tracknum_entry, FALSE);
        gtk_widget_set_sensitive(v2_comment_entry, FALSE);
        gtk_widget_set_sensitive(v2_composer_entry, FALSE);
        gtk_widget_set_sensitive(v2_orig_artist_entry, FALSE);
        gtk_widget_set_sensitive(v2_url_entry, FALSE);
        gtk_widget_set_sensitive(v2_encoded_by_entry, FALSE);
        gtk_widget_set_sensitive(v2_genre_combo, FALSE);
    } else {

        // If control reaches here, the toggle button is up
        // Enable labels
        for (i = 0; i < v2_labels_list->len; i++) {
            gtk_widget_set_sensitive(GTK_WIDGET( g_ptr_array_index(v2_labels_list, i) ), TRUE);
        }
        gtk_widget_set_sensitive(v2_title_entry, TRUE);
        gtk_widget_set_sensitive(v2_artist_entry, TRUE);
        gtk_widget_set_sensitive(v2_album_entry, TRUE);
        gtk_widget_set_sensitive(v2_year_entry, TRUE);
        gtk_widget_set_sensitive(v2_tracknum_entry, TRUE);
        gtk_widget_set_sensitive(v2_comment_entry, TRUE);
        gtk_widget_set_sensitive(v2_composer_entry, TRUE);
        gtk_widget_set_sensitive(v2_orig_artist_entry, TRUE);
        gtk_widget_set_sensitive(v2_url_entry, TRUE);
        gtk_widget_set_sensitive(v2_encoded_by_entry, TRUE);
        gtk_widget_set_sensitive(v2_genre_combo, TRUE);
    }
}
static void v1_genre_selected(GtkList *list, GtkWidget *w, gpointer data)
{
	void * p;
	p = gtk_object_get_data(GTK_OBJECT(w), "genre_id");
	if (p != NULL)
		v1_current_genre = GPOINTER_TO_INT(p);
	else
		v1_current_genre = 0;
}

static void v2_genre_selected(GtkList *list, GtkWidget *w, gpointer data)
{
	void * p;
	p = gtk_object_get_data(GTK_OBJECT(w), "genre_id");
	if (p != NULL)
		v2_current_genre = GPOINTER_TO_INT(p);
	else
		v2_current_genre = 0;
}

static void genre_set_popdown(GtkWidget *combo, GList *genres)
{
	struct genre_item *item;
	GtkWidget *w;

	while (genres)
	{
		item = genres->data;
		w = gtk_list_item_new_with_label(item->name);
		gtk_object_set_data(GTK_OBJECT(w), "genre_id",
				    GINT_TO_POINTER(item->id));
		gtk_widget_show(w);
		gtk_container_add(GTK_CONTAINER(GTK_COMBO(combo)->list), w);
		genres = genres->next;
	}
}

static void file_info_box_keypress_cb(GtkWidget *w, GdkEventKey *event, gpointer data)
{
	if (event && event->keyval == GDK_Escape)
		gtk_widget_destroy(w);
}

void mpg123_file_info_box(char *filename)
{
	int i;
	struct id3v1tag_t id3v1tag;
	FILE *fh;
	char *tmp, *title;
	const char *emphasis[4];
	const char *bool_label[2];

	emphasis[0] = _("None");
	emphasis[1] = _("50/15 ms");
	emphasis[2] = "";
	emphasis[3] = _("CCITT J.17");
	bool_label[0] = _("No");
	bool_label[1] = _("Yes");

	if (!window)
	{
		GtkWidget *window_vbox,
		          *id3v1_vbox, *id3v2_vbox, *id3v1_frame_vbox, *id3v2_frame_vbox,
		          *mpeg_lvbox, *mpeg_rvbox, *mpeg_hbox, *mpeg_box, *mpeg_frame,
		          *bbox, *save, *close, *copy_to, *copy_from,
		          *table, *label, *filename_hbox;

		v1_labels_list = g_ptr_array_new();
		v2_labels_list = g_ptr_array_new();

		window = gtk_window_new(GTK_WINDOW_DIALOG);
		gtk_signal_connect(GTK_OBJECT(window), "destroy", GTK_SIGNAL_FUNC(gtk_widget_destroyed), &window);
		gtk_signal_connect(GTK_OBJECT(window), "key_press_event", file_info_box_keypress_cb, NULL);
		gtk_container_set_border_width(GTK_CONTAINER(window), 10);

		window_vbox = gtk_vbox_new(FALSE,10);
		filename_hbox = gtk_hbox_new(FALSE, 5);
		gtk_box_pack_start(GTK_BOX(window_vbox), filename_hbox, FALSE, TRUE, 0);

		label = gtk_label_new(_("Filename:"));
		gtk_box_pack_start(GTK_BOX(filename_hbox), label, FALSE, TRUE, 0);
		filename_entry = xmms_entry_new();
		gtk_editable_set_editable(GTK_EDITABLE(filename_entry), FALSE);
		gtk_box_pack_start(GTK_BOX(filename_hbox), filename_entry, TRUE, TRUE, 0);

		/* Set up the genres list */

		if (!genre_list)
		{
			struct genre_item *item;

			for (i = 0; i < GENRE_MAX; i++)
			{
				item = g_malloc(sizeof (*item));
				item->name = gettext(mpg123_id3_genres[i]);
				item->id = i;
				genre_list = g_list_prepend(genre_list, item);
			}
			item = g_malloc(sizeof (*item));
			item->name = "";
			item->id = 0xff;
			genre_list = g_list_prepend(genre_list, item);
			genre_list = g_list_sort(genre_list, genre_comp_func);
		}

		notebook = gtk_notebook_new ();
		gtk_notebook_set_tab_pos (GTK_NOTEBOOK (notebook), GTK_POS_TOP);

		/* ID3v2 page */

		id3v2_vbox = gtk_vbox_new(FALSE, 0);

		id3v2_frame = gtk_frame_new("ID3v2 Information");
		gtk_box_pack_start(GTK_BOX(id3v2_vbox), id3v2_frame, FALSE, FALSE, 0);

		id3v2_frame_vbox = gtk_vbox_new(FALSE, 0);
		gtk_container_add(GTK_CONTAINER(id3v2_frame), id3v2_frame_vbox);

		table = gtk_table_new(6, 6, FALSE);
		gtk_container_set_border_width(GTK_CONTAINER(table), 5);
		gtk_widget_set_usize(GTK_WIDGET(table), 400, -1);
		gtk_box_pack_start(GTK_BOX(id3v2_frame_vbox), table, FALSE, FALSE, 0);

		v2_checkbox = gtk_check_button_new_with_label ("Disable ID3v2 Tag");
		gtk_signal_connect(GTK_OBJECT(v2_checkbox), "toggled", GTK_SIGNAL_FUNC(v2_toggle_cb), NULL);
		gtk_table_attach(GTK_TABLE(table), v2_checkbox, 1, 3, 0, 1, GTK_FILL, GTK_FILL, 0, 2);

		label = gtk_label_new("Track number:");
		g_ptr_array_add(v2_labels_list, (gpointer)label);
		gtk_misc_set_alignment(GTK_MISC(label), 1, 0.5);
		gtk_table_attach(GTK_TABLE(table), label, 3, 4, 0, 1,
			GTK_FILL, GTK_FILL, 5,5);

		v2_tracknum_entry = gtk_entry_new_with_max_length(3);
		gtk_widget_set_usize(v2_tracknum_entry, 20, -1);
		gtk_table_attach(GTK_TABLE(table), v2_tracknum_entry, 4, 5, 0, 1,
			GTK_FILL, GTK_FILL, 0, 2);

		label = gtk_label_new("Title:");
		g_ptr_array_add(v2_labels_list, (gpointer)label);
		gtk_misc_set_alignment(GTK_MISC(label), 1, 0.5);
		gtk_table_attach(GTK_TABLE(table), label, 0, 1, 1, 2, GTK_FILL, GTK_FILL, 5, 5);

		v2_title_entry = gtk_entry_new_with_max_length(MAX_ENTRY_LEN2);
		gtk_table_attach(GTK_TABLE(table), v2_title_entry, 1, 5, 1, 2,
			GTK_FILL | GTK_EXPAND | GTK_SHRINK,
			GTK_FILL | GTK_EXPAND | GTK_SHRINK, 0, 2);

		label = gtk_label_new("Artist:");
		g_ptr_array_add(v2_labels_list, (gpointer)label);
		gtk_misc_set_alignment(GTK_MISC(label), 1, 0.5);
		gtk_table_attach(GTK_TABLE(table), label, 0, 1, 2, 3,
			GTK_FILL, GTK_FILL, 5, 5);

		v2_artist_entry = gtk_entry_new_with_max_length(MAX_ENTRY_LEN2);
		gtk_table_attach(GTK_TABLE(table), v2_artist_entry, 1, 5, 2, 3,
			GTK_FILL | GTK_EXPAND | GTK_SHRINK,
			GTK_FILL | GTK_EXPAND | GTK_SHRINK, 0, 2);

		label = gtk_label_new("Album:");
		g_ptr_array_add(v2_labels_list, (gpointer)label);
		gtk_misc_set_alignment(GTK_MISC(label), 1, 0.5);
		gtk_table_attach(GTK_TABLE(table), label, 0, 1, 3, 4,
			GTK_FILL, GTK_FILL, 5, 5);

		v2_album_entry = gtk_entry_new_with_max_length(MAX_ENTRY_LEN2);
		gtk_table_attach(GTK_TABLE(table), v2_album_entry, 1, 5, 3, 4,
			GTK_FILL | GTK_EXPAND | GTK_SHRINK,
			GTK_FILL | GTK_EXPAND | GTK_SHRINK, 0, 2);

		label = gtk_label_new("Comment:");
		g_ptr_array_add(v2_labels_list, (gpointer)label);
		gtk_misc_set_alignment(GTK_MISC(label), 1, 0.5);
		gtk_table_attach(GTK_TABLE(table), label, 0, 1, 4, 5,
			GTK_FILL, GTK_FILL, 5, 5);

		v2_comment_entry = gtk_entry_new_with_max_length(MAX_ENTRY_LEN2);
		gtk_table_attach(GTK_TABLE(table), v2_comment_entry, 1, 5, 4, 5,
			GTK_FILL | GTK_EXPAND | GTK_SHRINK,
			GTK_FILL | GTK_EXPAND | GTK_SHRINK, 0, 2);

		label = gtk_label_new("Year:");
		g_ptr_array_add(v2_labels_list, (gpointer)label);
		gtk_misc_set_alignment(GTK_MISC(label), 1, 0.5);
		gtk_table_attach(GTK_TABLE(table), label, 0, 1, 5, 6,
			GTK_FILL, GTK_FILL, 5, 5);

		v2_year_entry = gtk_entry_new_with_max_length(4);
		gtk_widget_set_usize(v2_year_entry, 45, -1);
		gtk_table_attach(GTK_TABLE(table), v2_year_entry, 1, 2, 5, 6,
			GTK_FILL, GTK_FILL, 0, 2);

		label = gtk_label_new("Genre:");
		g_ptr_array_add(v2_labels_list, (gpointer)label);
		gtk_misc_set_alignment(GTK_MISC(label), 1, 0.5);
		gtk_table_attach(GTK_TABLE(table), label, 2, 3, 5, 6,
			GTK_FILL, GTK_FILL, 5, 5);

		v2_genre_combo = gtk_combo_new();
		gtk_combo_set_value_in_list(GTK_COMBO(v2_genre_combo), FALSE, TRUE);

		genre_set_popdown(v2_genre_combo, genre_list);
		gtk_signal_connect(GTK_OBJECT(GTK_COMBO(v2_genre_combo)->list),
				   "select-child", v2_genre_selected, NULL);

		gtk_table_attach(GTK_TABLE(table), v2_genre_combo, 3, 5, 5, 6,
			GTK_FILL | GTK_SHRINK, GTK_FILL |
			GTK_SHRINK, 0, 2);

		label = gtk_label_new("Composer:");
		g_ptr_array_add(v2_labels_list, (gpointer)label);
		gtk_misc_set_alignment(GTK_MISC(label), 1, 0.5);
		gtk_table_attach(GTK_TABLE(table), label, 0, 1, 6, 7,
			GTK_FILL, GTK_FILL, 5, 5);

		v2_composer_entry = gtk_entry_new_with_max_length(MAX_ENTRY_LEN2);
		gtk_table_attach(GTK_TABLE(table), v2_composer_entry, 1, 5, 6, 7,
			GTK_FILL | GTK_EXPAND | GTK_SHRINK,
			GTK_FILL | GTK_EXPAND | GTK_SHRINK, 0, 2);

		label = gtk_label_new("Orig. Artist:");
		g_ptr_array_add(v2_labels_list, (gpointer)label);
		gtk_misc_set_alignment(GTK_MISC(label), 1, 0.5);
		gtk_table_attach(GTK_TABLE(table), label, 0, 1, 7, 8,
			GTK_FILL, GTK_FILL, 5, 5);

		v2_orig_artist_entry = gtk_entry_new_with_max_length(MAX_ENTRY_LEN2);
		gtk_table_attach(GTK_TABLE(table), v2_orig_artist_entry, 1, 5, 7, 8,
			GTK_FILL | GTK_EXPAND | GTK_SHRINK,
			GTK_FILL | GTK_EXPAND | GTK_SHRINK, 0, 2);

		label = gtk_label_new("URL:");
		g_ptr_array_add(v2_labels_list, (gpointer)label);
		gtk_misc_set_alignment(GTK_MISC(label), 1, 0.5);
		gtk_table_attach(GTK_TABLE(table), label, 0, 1, 8, 9,
			GTK_FILL, GTK_FILL, 5, 5);

		v2_url_entry = gtk_entry_new_with_max_length(MAX_ENTRY_LEN2);
		gtk_table_attach(GTK_TABLE(table), v2_url_entry, 1, 5, 8, 9,
			GTK_FILL | GTK_EXPAND | GTK_SHRINK,
			GTK_FILL | GTK_EXPAND | GTK_SHRINK, 0, 2);
		label = gtk_label_new("Encoded By:");
		g_ptr_array_add(v2_labels_list, (gpointer)label);
		gtk_misc_set_alignment(GTK_MISC(label), 1, 0.5);
		gtk_table_attach(GTK_TABLE(table), label, 0, 1, 9, 10,
			GTK_FILL, GTK_FILL, 5, 5);

		v2_encoded_by_entry = gtk_entry_new_with_max_length(MAX_ENTRY_LEN2);
		gtk_table_attach(GTK_TABLE(table), v2_encoded_by_entry, 1, 5, 9, 10,
			GTK_FILL | GTK_EXPAND | GTK_SHRINK,
			GTK_FILL | GTK_EXPAND | GTK_SHRINK, 0, 2);

		bbox = gtk_hbutton_box_new();
		gtk_button_box_set_layout (GTK_BUTTON_BOX (bbox), GTK_BUTTONBOX_START);
		gtk_button_box_set_spacing (GTK_BUTTON_BOX (bbox), 0);
		gtk_container_set_border_width(GTK_CONTAINER(bbox), 5);
		gtk_box_pack_start(GTK_BOX(id3v2_frame_vbox), bbox, FALSE, FALSE, 0);

		copy_to = gtk_button_new_with_label("ID3v2 to ID3v1");
		gtk_signal_connect(GTK_OBJECT(copy_to), "clicked", GTK_SIGNAL_FUNC(copy_v2_to_v1_cb), NULL);
		/* remove the next line to thicken the button width */
		GTK_WIDGET_SET_FLAGS(copy_to, GTK_CAN_DEFAULT);
		gtk_box_pack_start(GTK_BOX(bbox), copy_to, FALSE, TRUE, 0);

		label = gtk_label_new ("ID3v2");
		gtk_notebook_append_page (GTK_NOTEBOOK (notebook), id3v2_vbox, label);

		/* ID3v1 page */

		id3v1_vbox = gtk_vbox_new(FALSE, 10);
		id3v1_frame = gtk_frame_new("ID3v1 Information");
		gtk_box_pack_start(GTK_BOX(id3v1_vbox), id3v1_frame, TRUE, TRUE, 0);

		id3v1_frame_vbox = gtk_vbox_new(FALSE,10);
		gtk_container_add(GTK_CONTAINER(id3v1_frame), id3v1_frame_vbox);

		table = gtk_table_new(6, 6, FALSE);
		gtk_container_set_border_width(GTK_CONTAINER(table), 5);
		//gtk_widget_set_usize(GTK_WIDGET(table), 325, -1);
		//gtk_container_add(GTK_CONTAINER(id3v1_frame), table);
		gtk_box_pack_start(GTK_BOX(id3v1_frame_vbox), table, FALSE, FALSE, 0);

		v1_checkbox = gtk_check_button_new_with_label ("Disable ID3v1 Tag");
		gtk_signal_connect(GTK_OBJECT(v1_checkbox), "toggled", GTK_SIGNAL_FUNC(v1_toggle_cb), NULL);
		gtk_table_attach(GTK_TABLE(table), v1_checkbox, 1, 3, 0, 1,
		GTK_FILL, GTK_FILL, 0, 2);

		label = gtk_label_new("Track number:");
		g_ptr_array_add(v1_labels_list, (gpointer)label);
		gtk_misc_set_alignment(GTK_MISC(label), 1, 0.5);
		gtk_table_attach(GTK_TABLE(table), label, 3, 4, 0, 1,
		GTK_FILL, GTK_FILL, 5,5);

		v1_tracknum_entry = gtk_entry_new_with_max_length(3);
		gtk_widget_set_usize(v1_tracknum_entry, 20, -1);
		gtk_table_attach(GTK_TABLE(table), v1_tracknum_entry, 4, 5, 0, 1,
		GTK_FILL, GTK_FILL, 0, 2);

		label = gtk_label_new("Title:");
		g_ptr_array_add(v1_labels_list, (gpointer)label);
		gtk_misc_set_alignment(GTK_MISC(label), 1, 0.5);
		gtk_table_attach(GTK_TABLE(table), label, 0, 1, 1, 2,
		GTK_FILL, GTK_FILL, 5, 5);

		v1_title_entry = gtk_entry_new_with_max_length(30);
		gtk_table_attach(GTK_TABLE(table), v1_title_entry, 1, 5, 1, 2,
		GTK_FILL | GTK_EXPAND | GTK_SHRINK,
		GTK_FILL | GTK_EXPAND | GTK_SHRINK, 0, 2);

		label = gtk_label_new("Artist:");
		g_ptr_array_add(v1_labels_list, (gpointer)label);
		gtk_misc_set_alignment(GTK_MISC(label), 1, 0.5);
		gtk_table_attach(GTK_TABLE(table), label, 0, 1, 2, 3,
		GTK_FILL, GTK_FILL, 5, 5);

		v1_artist_entry = gtk_entry_new_with_max_length(30);
		gtk_table_attach(GTK_TABLE(table), v1_artist_entry, 1, 5, 2, 3,
		GTK_FILL | GTK_EXPAND | GTK_SHRINK,
		GTK_FILL | GTK_EXPAND | GTK_SHRINK, 0, 2);

		label = gtk_label_new("Album:");
		g_ptr_array_add(v1_labels_list, (gpointer)label);
		gtk_misc_set_alignment(GTK_MISC(label), 1, 0.5);
		gtk_table_attach(GTK_TABLE(table), label, 0, 1, 3, 4,
		GTK_FILL, GTK_FILL, 5, 5);

		v1_album_entry = gtk_entry_new_with_max_length(30);
		gtk_table_attach(GTK_TABLE(table), v1_album_entry, 1, 5, 3, 4,
		GTK_FILL | GTK_EXPAND | GTK_SHRINK,
		GTK_FILL | GTK_EXPAND | GTK_SHRINK, 0, 2);

		label = gtk_label_new("Comment:");
		g_ptr_array_add(v1_labels_list, (gpointer)label);
		gtk_misc_set_alignment(GTK_MISC(label), 1, 0.5);
		gtk_table_attach(GTK_TABLE(table), label, 0, 1, 4, 5,
		GTK_FILL, GTK_FILL, 5, 5);

		v1_comment_entry = gtk_entry_new_with_max_length(30);
		gtk_table_attach(GTK_TABLE(table), v1_comment_entry, 1, 5, 4, 5,
		GTK_FILL | GTK_EXPAND | GTK_SHRINK,
		GTK_FILL | GTK_EXPAND | GTK_SHRINK, 0, 2);

		label = gtk_label_new("Year:");
		g_ptr_array_add(v1_labels_list, (gpointer)label);
		gtk_misc_set_alignment(GTK_MISC(label), 1, 0.5);
		gtk_table_attach(GTK_TABLE(table), label, 0, 1, 5, 6,
		GTK_FILL, GTK_FILL, 5, 5);

		v1_year_entry = gtk_entry_new_with_max_length(4);
		gtk_widget_set_usize(v1_year_entry, 45, -1);
		gtk_table_attach(GTK_TABLE(table), v1_year_entry, 1, 2, 5, 6,
		GTK_FILL, GTK_FILL, 0, 2);

		label = gtk_label_new("Genre:");
		g_ptr_array_add(v1_labels_list, (gpointer)label);
		gtk_misc_set_alignment(GTK_MISC(label), 1, 0.5);
		gtk_table_attach(GTK_TABLE(table), label, 2, 3, 5, 6,
		GTK_FILL, GTK_FILL, 5, 5);

		v1_genre_combo = gtk_combo_new();
		gtk_entry_set_editable(GTK_ENTRY(GTK_COMBO(v1_genre_combo)->entry), FALSE);
		genre_set_popdown(v1_genre_combo, genre_list);
		gtk_signal_connect(GTK_OBJECT(GTK_COMBO(v1_genre_combo)->list),
			   "select-child", v1_genre_selected, NULL);

		gtk_table_attach(GTK_TABLE(table), v1_genre_combo, 3, 5, 5, 6,
			GTK_FILL | GTK_SHRINK, GTK_FILL |
			GTK_SHRINK, 0, 2);

		bbox = gtk_hbutton_box_new();
		gtk_button_box_set_layout (GTK_BUTTON_BOX (bbox), GTK_BUTTONBOX_START);
		gtk_button_box_set_spacing (GTK_BUTTON_BOX (bbox), 0);
		gtk_container_set_border_width(GTK_CONTAINER(bbox), 5);
		gtk_box_pack_start(GTK_BOX(id3v1_frame_vbox), bbox, FALSE, FALSE, 0);

		copy_from = gtk_button_new_with_label("ID3v1 to ID3v2");
		gtk_signal_connect(GTK_OBJECT(copy_from), "clicked", GTK_SIGNAL_FUNC(copy_v1_to_v2_cb), NULL);
		// remove the next line to thicken the button width
		GTK_WIDGET_SET_FLAGS(copy_from, GTK_CAN_DEFAULT);
		gtk_box_pack_start(GTK_BOX(bbox), copy_from, FALSE, TRUE, 0);

		label = gtk_label_new ("ID3v1");
		gtk_notebook_append_page (GTK_NOTEBOOK (notebook), id3v1_vbox, label);

		/* MPEG info page */

		mpeg_frame = gtk_frame_new("MPEG Information");
		mpeg_hbox = gtk_hbox_new(FALSE,50);
		gtk_container_add(GTK_CONTAINER(mpeg_frame), mpeg_hbox);

		mpeg_lvbox = gtk_vbox_new(FALSE, 5);
		gtk_container_set_border_width(GTK_CONTAINER(mpeg_lvbox), 10);
		gtk_box_pack_start(GTK_BOX(mpeg_hbox), mpeg_lvbox, FALSE, FALSE, 0);

		mpeg_box = gtk_vbox_new(FALSE, 5);
		gtk_box_pack_start(GTK_BOX(mpeg_hbox), mpeg_box, FALSE, FALSE, 0);
		gtk_container_set_border_width(GTK_CONTAINER(mpeg_box), 10);
		gtk_box_set_spacing(GTK_BOX(mpeg_box), 0);

		mpeg_level = gtk_label_new("");
		//gtk_widget_set_usize(mpeg_level, 120, -2);
		gtk_label_set_justify (GTK_LABEL(mpeg_level), GTK_JUSTIFY_LEFT);
		gtk_misc_set_alignment(GTK_MISC(mpeg_level), 0, 0.5);
		gtk_box_pack_start(GTK_BOX(mpeg_lvbox), mpeg_level, FALSE, FALSE, 0);

		mpeg_samplerate = gtk_label_new("");
		gtk_label_set_justify (GTK_LABEL(mpeg_samplerate), GTK_JUSTIFY_LEFT);
		gtk_misc_set_alignment(GTK_MISC(mpeg_samplerate), 0, 0.5);
		gtk_box_pack_start(GTK_BOX(mpeg_lvbox), mpeg_samplerate, FALSE, FALSE, 0);

		mpeg_fileinfo = gtk_label_new("");
		gtk_label_set_justify (GTK_LABEL(mpeg_fileinfo), GTK_JUSTIFY_LEFT);
		gtk_misc_set_alignment(GTK_MISC(mpeg_fileinfo), 0, 0.5);
		gtk_box_pack_start(GTK_BOX(mpeg_lvbox), mpeg_fileinfo, FALSE, FALSE, 0);

		mpeg_rvbox = gtk_vbox_new(FALSE, 5);
		gtk_box_pack_start(GTK_BOX(mpeg_hbox), mpeg_rvbox, FALSE, FALSE, 0);
		gtk_container_set_border_width(GTK_CONTAINER(mpeg_rvbox), 10);

		mpeg_bitrate = gtk_label_new("");
		gtk_label_set_justify (GTK_LABEL(mpeg_bitrate), GTK_JUSTIFY_LEFT);
		gtk_misc_set_alignment(GTK_MISC(mpeg_bitrate), 0, 0.5);
		gtk_box_pack_start(GTK_BOX(mpeg_rvbox), mpeg_bitrate, FALSE, FALSE, 0);

		mpeg_flags = gtk_label_new("");
		gtk_label_set_justify (GTK_LABEL(mpeg_flags), GTK_JUSTIFY_LEFT);
		gtk_misc_set_alignment(GTK_MISC(mpeg_flags), 0, 0.5);
		gtk_box_pack_start(GTK_BOX(mpeg_rvbox), mpeg_flags, FALSE, FALSE, 0);

		label = gtk_label_new ("MPEG");
		gtk_notebook_append_page (GTK_NOTEBOOK (notebook), mpeg_frame, label);

		/* add notebook to window vbox */
		gtk_box_pack_start(GTK_BOX(window_vbox), notebook, FALSE, FALSE, 0);

		/* add button box to window vbox */
		bbox = gtk_hbutton_box_new();
		gtk_box_pack_start(GTK_BOX(window_vbox), bbox, FALSE, FALSE, 0);

		gtk_button_box_set_layout (GTK_BUTTON_BOX (bbox), GTK_BUTTONBOX_END);
		gtk_button_box_set_spacing (GTK_BUTTON_BOX (bbox), 5);

		save = gtk_button_new_with_label("Save");
		gtk_signal_connect(GTK_OBJECT(save), "clicked", GTK_SIGNAL_FUNC(save_cb), NULL);
		gtk_box_pack_start(GTK_BOX(bbox), save, TRUE, TRUE, 5);

		close = gtk_button_new_with_label("Close");
		gtk_signal_connect_object(GTK_OBJECT(close), "clicked",
		GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(window));
		gtk_box_pack_start(GTK_BOX(bbox), close, TRUE, TRUE, 5);

		gtk_container_add(GTK_CONTAINER(window), window_vbox);
		gtk_widget_show_all(window);
	}

	if (current_filename)
		g_free(current_filename);
	current_filename = g_strdup(filename);

	title = g_strdup_printf(_("File Info - %s"), g_basename(filename));
	gtk_window_set_title(GTK_WINDOW(window), title);
	g_free(title);

	gtk_entry_set_text(GTK_ENTRY(filename_entry), filename);
	gtk_editable_set_position(GTK_EDITABLE(filename_entry), -1);

	title = g_strdup(g_basename(filename));
	if ((tmp = strrchr(title, '.')) != NULL)
		*tmp = '\0';
	gtk_entry_set_text(GTK_ENTRY(v1_title_entry), title);
	gtk_entry_set_text(GTK_ENTRY(v2_title_entry), title);
	g_free(title);

	gtk_entry_set_text(GTK_ENTRY(v1_artist_entry), "");
	gtk_entry_set_text(GTK_ENTRY(v1_album_entry), "");
	gtk_entry_set_text(GTK_ENTRY(v1_year_entry), "");
	gtk_entry_set_text(GTK_ENTRY(v1_tracknum_entry), "");
	gtk_entry_set_text(GTK_ENTRY(v1_comment_entry), "");

	gtk_list_select_item(GTK_LIST(GTK_COMBO(v1_genre_combo)->list),
			     genre_find_index(genre_list, 0xff));

	gtk_entry_set_text(GTK_ENTRY(v2_artist_entry), "");
	gtk_entry_set_text(GTK_ENTRY(v2_album_entry), "");
	gtk_entry_set_text(GTK_ENTRY(v2_year_entry), "");
	gtk_entry_set_text(GTK_ENTRY(v2_tracknum_entry), "");
	gtk_entry_set_text(GTK_ENTRY(v2_comment_entry), "");
	gtk_entry_set_text(GTK_ENTRY(v2_composer_entry), "");
	gtk_entry_set_text(GTK_ENTRY(v2_orig_artist_entry), "");
	gtk_entry_set_text(GTK_ENTRY(v2_url_entry), "");
	gtk_entry_set_text(GTK_ENTRY(v2_encoded_by_entry), "");
	gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(v2_genre_combo)->entry), "");

	gtk_label_set_text(GTK_LABEL(mpeg_level), "MPEG ?, layer ?");
	gtk_label_set_text(GTK_LABEL(mpeg_bitrate), "");
	gtk_label_set_text(GTK_LABEL(mpeg_samplerate), "");
	gtk_label_set_text(GTK_LABEL(mpeg_flags), "");
	gtk_label_set_text(GTK_LABEL(mpeg_fileinfo), "");
	if (!strncasecmp(filename, "http://", 7))
	{
		file_info_http(filename);
		return;
	}

	gtk_widget_set_sensitive(id3v1_frame, TRUE);
	gtk_widget_set_sensitive(id3v2_frame, TRUE);

	if ((fh = fopen(current_filename, "rb")) != NULL)
	{
		struct frame frm;
		gboolean id3_found = FALSE;
		struct id3_tag *id3 = NULL;
		guint8 *buf;
		double tpf;
		int pos;
		xing_header_t xing_header;
		guint32 num_frames;
		/*
		 * Try reading ID3v2 tag.
		 */
		fseek(fh, 0, SEEK_SET);
		id3 = id3_open_fp(fh, 0);
		if (id3 != NULL)
		{
			struct id3_frame* frame;

			set_entry_tag_v2(GTK_ENTRY(v2_title_entry), id3, ID3_TIT2);
			set_entry_tag_v2(GTK_ENTRY(v2_artist_entry), id3, ID3_TPE1);
			set_entry_tag_v2(GTK_ENTRY(v2_album_entry), id3, ID3_TALB);
			set_entry_tag_v2(GTK_ENTRY(v2_comment_entry), id3, ID3_COMM);
			set_entry_tag_v2(GTK_ENTRY(v2_composer_entry), id3, ID3_TCOM);
			set_entry_tag_v2(GTK_ENTRY(v2_orig_artist_entry), id3, ID3_TOPE);
			set_entry_tag_v2(GTK_ENTRY(v2_url_entry), id3, ID3_WCOM);
			set_entry_tag_v2(GTK_ENTRY(v2_encoded_by_entry), id3, ID3_TENC);
			set_entry_tag_v2(GTK_ENTRY(v2_tracknum_entry), id3, ID3_TRCK);
			set_entry_tag_v2(GTK_ENTRY(v2_year_entry), id3, ID3_TYER);

			frame = id3_get_frame(id3, ID3_TCON, 1);
			if (frame != NULL)
			{
				char* genre = id3_get_content(frame);
				if (genre != NULL)
				{
					int genre_idx = genre_find_index_str(genre_list, genre);
					if (genre_idx > 0)
						gtk_list_select_item(GTK_LIST(GTK_COMBO(v2_genre_combo)->list), genre_idx);
					else
						gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(v2_genre_combo)->entry), genre);
				}
			}

			id3_close(id3);
		}
		else
		{
			/* Grey out the id3v2 tab */
			gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(v2_checkbox), TRUE);
		}

		/*
		* Try reading ID3v1 tag.
		*/
		fseek(fh, -sizeof (id3v1tag), SEEK_END);
		if ( (fread(&id3v1tag, 1, sizeof (id3v1tag), fh) == sizeof (id3v1tag)) && !strncmp(id3v1tag.tag, "TAG", 3))
		{
		id3_found = TRUE;
		set_entry_tag_v1(GTK_ENTRY(v1_title_entry), id3v1tag.title, 30);
		set_entry_tag_v1(GTK_ENTRY(v1_artist_entry), id3v1tag.artist, 30);
		set_entry_tag_v1(GTK_ENTRY(v1_album_entry), id3v1tag.album, 30);
		set_entry_tag_v1(GTK_ENTRY(v1_year_entry), id3v1tag.year, 4);
		/* Check for v1.1 tags */
		if (id3v1tag.u.v1_1.__zero == 0 && id3v1tag.u.v1_1.track_number > 0)
		{
			char *temp = g_strdup_printf("%d", id3v1tag.u.v1_1.track_number);
			set_entry_tag_v1(GTK_ENTRY(v1_comment_entry), id3v1tag.u.v1_1.comment, 28);
			gtk_entry_set_text(GTK_ENTRY(v1_tracknum_entry), temp);
			g_free(temp);
		}
		else
		{
			set_entry_tag_v1(GTK_ENTRY(v1_comment_entry), id3v1tag.u.v1_0.comment, 30);
			gtk_entry_set_text(GTK_ENTRY(v1_tracknum_entry), "");
		}
		gtk_list_select_item(GTK_LIST(GTK_COMBO(v1_genre_combo)->list), genre_find_index(genre_list, id3v1tag.genre));
		}
		else
		{
		// Grey out id3v1 tab
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(v1_checkbox), TRUE);
		}

		rewind(fh);
		
		if (!mpg123_get_first_frame(fh, &frm, &buf))
		{
			fclose(fh);
			return;
		}

		tpf = mpg123_compute_tpf(&frm);
		set_mpeg_level_label(frm.mpeg25, frm.lsf, frm.lay);
		pos = ftell(fh);
		fseek(fh, 0, SEEK_END);
		if (mpg123_get_xing_header(&xing_header, buf))
		{
			num_frames = xing_header.frames;
			label_set_text(mpeg_bitrate,
				       _("Bitrate: Variable,\n"
					 "avg. bitrate: %d kb/s"),
				       (int)((xing_header.bytes * 8) /
					     (tpf * xing_header.frames * 1000)));
		}
		else
		{
			num_frames = ((ftell(fh) - pos - (id3_found ? 128 : 0)) /
				      mpg123_compute_bpf(&frm)) + 1;
			label_set_text(mpeg_bitrate,
				       _("Bitrate: %d kb/s"),
				       tabsel_123[frm.lsf][frm.lay - 1][frm.bitrate_index]);
		}
		label_set_text(mpeg_samplerate, _("Samplerate: %d Hz"),
			       mpg123_freqs[frm.sampling_frequency]);
		label_set_text(mpeg_flags, _("%s\nError protection: %s\n"
					     "Copyright: %s\nOriginal: %s\n"
					     "Emphasis: %s"),
			       channel_mode_name(frm.mode),
			       bool_label[frm.error_protection],
			       bool_label[frm.copyright], bool_label[frm.original],
			       emphasis[frm.emphasis]);
		label_set_text(mpeg_fileinfo, _("%d frames\nFilesize: %lu B"),
			       num_frames, ftell(fh));
		g_free(buf);
		fclose(fh);
	}
}
