/*
 *    Copyright (C) 1999, 2001,  Espen Skoglund
 *    Copyright (C) 2000 - 2004  Haavard Kvaalen
 *
 * $Id: id3.c,v 1.14 2004/04/04 16:14:31 havard Exp $
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
#include "config.h"

#include <sys/types.h>
#include <sys/uio.h>
#include <sys/stat.h>
#include <glib.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>

#include "id3.h"
#include "id3_header.h"


/*
**
** Functions for accessing the ID3 tag using a memory pointer.
**
*/

/*
 * Function id3_seek_mem (id3, offset)
 *
 *    Seek `offset' bytes forward in the indicated ID3-tag.  Return 0
 *    upon success, or -1 if an error occured.
 *
 */
static int id3_seek_mem(struct id3_tag *id3, int offset)
{
	if (id3->id3_pos + offset > id3->id3_totalsize ||
	    id3->id3_pos + offset < 0)
	{
		id3_error(id3, "seeking beyond tag boundary");
		return -1;
	}
	id3->s.me.id3_ptr = (char *) id3->s.me.id3_ptr + offset;
	id3->id3_pos += offset;

	return 0;
}


/*
 * Function id3_read_mem (id3, buf, size)
 *
 *    Read `size' bytes from indicated ID3-tag.  If `buf' is non-NULL,
 *    read into that buffer.  Return a pointer to the data which was
 *    read, or NULL upon error.
 *
 */
static void *id3_read_mem(struct id3_tag *id3, void *buf, int size)
{
	void *ret = id3->s.me.id3_ptr;

	/*
	 * Check boundary.
	 */
	if (id3->id3_pos + size > id3->id3_totalsize)
		return NULL;

	/*
	 * If buffer is non-NULL, we have to copy the data.
	 */
	if (buf != NULL)
	{
		if (size > ID3_FD_BUFSIZE)
			return NULL;
		memcpy(buf, id3->s.me.id3_ptr, size);
	}

	/*
	 * Update memory pointer.
	 */
	id3->s.me.id3_ptr = (char *) id3->s.me.id3_ptr + size;
	id3->id3_pos += size;

	return ret;
}


/*
**
** Functions for accessing the ID3 tag using a file descriptor.
**
*/

/*
 * Function id3_seek_fd (id3, offset)
 *
 *    Seek `offset' bytes forward in the indicated ID3-tag.  Return 0
 *    upon success, or -1 if an error occured.
 *
 */
static int id3_seek_fd(struct id3_tag *id3, int offset)
{
	/*
	 * Check boundary.
	 */
	if (id3->id3_pos + offset > id3->id3_totalsize ||
	    id3->id3_pos + offset < 0)
		return -1;

	if (lseek(id3->s.fd.id3_fd, offset, SEEK_CUR) == -1)
	{
		id3_error(id3, "seeking beyond tag boundary");
		return -1;
	}
	id3->id3_pos += offset;

	return 0;
}


/*
 * Function id3_read_fd (id3, buf, size)
 *
 *    Read `size' bytes from indicated ID3-tag.  If `buf' is non-NULL,
 *    read into that buffer.  Return a pointer to the data which was
 *    read, or NULL upon error.
 *
 */
static void *id3_read_fd(struct id3_tag *id3, void *buf, int size)
{
	int done = 0;

	/*
	 * Check boundary.
	 */
	if (id3->id3_pos + size > id3->id3_totalsize)
		size = id3->id3_totalsize - id3->id3_pos;

	/*
	 * If buffer is NULL, we use the default buffer.
	 */
	if (buf == NULL)
	{
		if (size > ID3_FD_BUFSIZE)
			return NULL;
		buf = id3->s.fd.id3_buf;
	}

	/*
	 * Read until we have slurped as much data as we wanted.
	 */
	while (done < size)
	{
		char *buffer = (char *)buf + done;
		int ret;

		/*
		 * Try reading from file.
		 */
		ret = read(id3->s.fd.id3_fd, buffer, size);
		if (ret <= 0)
		{
			id3_error(id3, "read(2) failed");
			return NULL;
		}

		id3->id3_pos += ret;
		done += ret;
	}

	return buf;
}


/*
**
** Functions for accessing the ID3 tag using a file pointer.
**
*/

/*
 * Function id3_seek_fp (id3, offset)
 *
 *    Seek `offset' bytes forward in the indicated ID3-tag.  Return 0
 *    upon success, or -1 if an error occured.
 *
 */
static int id3_seek_fp(struct id3_tag *id3, int offset)
{
	/*
	 * Check boundary.
	 */
	if (id3->id3_pos + offset > id3->id3_totalsize ||
	    id3->id3_pos + offset < 0)
		return -1;

	if (offset > 0)
	{
		/*
		 * If offset is positive, we use fread() instead of fseek().  This
		 * is more robust with respect to streams.
		 */
		char buf[64];
		int r, remain = offset;

		while (remain > 0)
		{
			int size = MIN(64, remain);
			r = fread(buf, 1, size, id3->s.fp.id3_fp);
			if (r == 0)
			{
				id3_error(id3, "fread() failed");
				return -1;
			}
			remain -= r;
		}
	}
	else
	{
		/*
		 * If offset is negative, we ahve to use fseek().  Let us hope
		 * that it works.
		 */
		if (fseek(id3->s.fp.id3_fp, offset, SEEK_CUR) == -1)
		{
			id3_error(id3, "seeking beyond tag boundary");
			return -1;
		}
	}
	id3->id3_pos += offset;

	return 0;
}


/*
 * Function id3_read_fp (id3, buf, size)
 *
 *    Read `size' bytes from indicated ID3-tag.  If `buf' is non-NULL,
 *    read into that buffer.  Return a pointer to the data which was
 *    read, or NULL upon error.
 *
 */
static void *id3_read_fp(struct id3_tag *id3, void *buf, int size)
{
	int ret;

	/*
	 * Check boundary.
	 */
	if (id3->id3_pos + size > id3->id3_totalsize)
		size = id3->id3_totalsize - id3->id3_pos;

	/*
	 * If buffer is NULL, we use the default buffer.
	 */
	if (buf == NULL)
	{
		if (size > ID3_FD_BUFSIZE)
			return NULL;
		buf = id3->s.fd.id3_buf;
	}

	/*
	 * Try reading from file.
	 */
	ret = fread(buf, 1, size, id3->s.fp.id3_fp);
	if (ret != size)
	{
		id3_error(id3, "fread() failed");
		return NULL;
	}

	id3->id3_pos += ret;

	return buf;
}




/*
 * Function id3_open_mem (ptr, flags)
 *
 *    Open an ID3 tag using a memory pointer.  Return a pointer to a
 *    structure describing the ID3 tag, or NULL if an error occured.
 *
 */
struct id3_tag *id3_open_mem(void *ptr, int flags)
{
	struct id3_tag *id3;

	/*
	 * Allocate ID3 structure.
	 */
	id3 = g_malloc0(sizeof (struct id3_tag));

	/*
	 * Initialize access pointers.
	 */
	id3->id3_seek = id3_seek_mem;
	id3->id3_read = id3_read_mem;

	id3->id3_oflags = flags;
	id3->id3_type = ID3_TYPE_MEM;
	id3->id3_pos = 0;
	id3->s.me.id3_ptr = ptr;

	/*
	 * Try reading ID3 tag.
	 */
	if (id3_read_tag(id3) == -1)
	{
		if (~flags & ID3_OPENF_CREATE)
			goto Return_NULL;
		id3_init_tag(id3);
	}

	return id3;

 Return_NULL:
	g_free(id3);
	return NULL;
}

/*
 * Function id3_new()
 *
 *    Creates a new ID3 tag structure. Useful for creating
 *    a new tag.
 *
 */
struct id3_tag *id3_new()
{
    struct id3_tag *id3;

    /*
     * Allocate ID3 structure.
     */
    id3 = g_malloc0( sizeof(struct id3_tag) );
    
    if (id3 != NULL)
    {
       id3_init_tag ( id3 );
       id3->id3_type = ID3_TYPE_NONE;
       id3->id3_seek = NULL;
       id3->id3_read = NULL;
    }

    return id3;
}

/*
 * Function id3_open_fd (fd, flags)
 *
 *    Open an ID3 tag using a file descriptor.  Return a pointer to a
 *    structure describing the ID3 tag, or NULL if an error occured.
 *
 */
struct id3_tag *id3_open_fd(int fd, int flags)
{
	struct id3_tag *id3;

	/*
	 * Allocate ID3 structure.
	 */
	id3 = g_malloc0(sizeof(struct id3_tag));

	/*
	 * Initialize access pointers.
	 */
	id3->id3_seek = id3_seek_fd;
	id3->id3_read = id3_read_fd;

	id3->id3_oflags = flags;
	id3->id3_type = ID3_TYPE_FD;
	id3->id3_pos = 0;
	id3->s.fd.id3_fd = fd;

	/*
	 * Allocate buffer to hold read data.
	 */
	id3->s.fd.id3_buf = g_malloc(ID3_FD_BUFSIZE);

	/*
	 * Try reading ID3 tag.
	 */
	if (id3_read_tag(id3) == -1)
	{
		if (~flags & ID3_OPENF_CREATE)
			goto Return_NULL;
		id3_init_tag(id3);
	}

	return id3;

	/*
	 * Cleanup code.
	 */
 Return_NULL:
	g_free(id3->s.fd.id3_buf);
	g_free(id3);
	return NULL;
}


/*
 * Function id3_open_fp (fp, flags)
 *
 *    Open an ID3 tag using a file pointer.  Return a pointer to a
 *    structure describing the ID3 tag, or NULL if an error occured.
 *
 */
struct id3_tag *id3_open_fp(FILE *fp, int flags)
{
	struct id3_tag *id3;

	/*
	 * Allocate ID3 structure.
	 */
	id3 = g_malloc0(sizeof(struct id3_tag));

	/*
	 * Initialize access pointers.
	 */
	id3->id3_seek = id3_seek_fp;
	id3->id3_read = id3_read_fp;

	id3->id3_oflags = flags;
	id3->id3_type = ID3_TYPE_FP;
	id3->id3_pos = 0;
	id3->s.fp.id3_fp = fp;

	/*
	 * Allocate buffer to hold read data.
	 */
	id3->s.fp.id3_buf = g_malloc(ID3_FD_BUFSIZE);

	/*
	 * Try reading ID3 tag.
	 */
	if (id3_read_tag(id3) == -1)
	{
		if (~flags & ID3_OPENF_CREATE)
			goto Return_NULL;
		id3_init_tag(id3);
	}

	return id3;

	/*
	 * Cleanup code.
	 */
 Return_NULL:
	g_free(id3->s.fp.id3_buf);
	g_free(id3);
	return NULL;
}


/*
 * Function id3_close (id3)
 *
 *    Free all resources assoicated with the ID3 tag.
 *
 */
int id3_close(struct id3_tag *id3)
{
	int ret = 0;

	switch(id3->id3_type)
	{
		case ID3_TYPE_MEM:
			break;
		case ID3_TYPE_FD:
			g_free(id3->s.fd.id3_buf);
			break;
		case ID3_TYPE_FP:
			g_free(id3->s.fp.id3_buf);
			break;
		case ID3_TYPE_NONE:
			id3_error(id3, "unknown ID3 type");
			ret = -1;
	}

	id3_destroy_frames(id3);

	g_free(id3);

	return ret;
}


/*
 * Function id3_tell (id3)
 *
 *    Return the current position in ID3 tag.  This will always be
 *    directly after the tag.
 *
 */
#if 0
int id3_tell(struct id3_tag *id3)
{
	if (id3->id3_newtag)
		return 0;
	else
		return id3->id3_totalsize + 3 + sizeof(id3_taghdr_t);
}
#endif


/*
 * Function id3_alter_file (id3)
 *
 *    When altering a file, some ID3 tags should be discarded.  As the ID3
 *    library has no means of knowing when a file has been altered
 *    outside of the library, this function must be called manually
 *    whenever the file is altered.
 *
 */
int id3_alter_file(struct id3_tag *id3)
{
	/*
	 * List of frame classes that should be discarded whenever the
	 * file is altered.
	 */
	static guint32 discard_list[] = {
		ID3_ETCO, ID3_EQUA, ID3_MLLT, ID3_POSS, ID3_SYLT,
		ID3_SYTC, ID3_RVAD, ID3_TENC, ID3_TLEN, ID3_TSIZ,
		0
	};
	struct id3_frame *fr;
	guint32 id, i = 0;

	/*
	 * Go through list of frame types that should be discarded.
	 */
	while ((id = discard_list[i++]) != 0)
	{
		/*
		 * Discard all frames of that type.
		 */
		while ((fr = id3_get_frame(id3, id, 1)))
			id3_delete_frame(fr);
	}

	return 0;
}


/*
 * Function safe_write (fd, buf, size)
 *
 *    Like write(2), except that the whole buffer will be written.
 *
 */
static int safe_write(int fd, void *buf, int size)
{
	int remaining = size;
	char *ptr = buf;
	int r;

	while (remaining > 0)
	{
		if ((r = write(fd, ptr, remaining)) == -1)
			return -1;
		remaining -= r;
		ptr += r;
	}

	return 0;
}

/*
 * Function id3_remove_tag_filename(filename)
 *
 *   Remove the ID3v2 tag from the file indicated by filename. Takes care of resizing
 *   the file, if needed. Returns 0 upon success, or -1 if an error occured.
 *
 */
int id3_remove_tag_filename(const char* filename)
{
    struct id3_tag *current_id3; 
    int fd;
    int current_totalsize;
    struct stat stat_buf;
    
    fd = open(filename, O_RDWR);
    if (fd == -1)
       return -1;
      
    /*
     * Figure out how large the current tag is.
     */
    current_id3 = id3_open_fd(fd, 0);
    if (current_id3 != NULL)
    {
       /* We use MAX to make sure an erroneous tag doesn't confuse us */
       current_totalsize = MAX(current_id3->id3_totalsize, 0);
       id3_close(current_id3);
    }
    else
    {
       current_totalsize = 0;
    }
    
    if (current_totalsize <= 0)
       return 0;
    
    /*
     * Rewrite the file.
     */

    stat(filename, &stat_buf); 

    /* Re-position all the data current_totalsize bytes backwards */
    {
          int read_pos, write_pos;
          size_t read_size;
          char buf[4096];

          /* TODO: Add error handling to IO operations */
          
          /* We reposition the data by going forwards, each time reading
             a block and copying it backward. That way, we never write
             over a block which we intend to read later. */
          write_pos = 0;
          read_pos = write_pos + current_totalsize;

          do
          {
             lseek(fd, read_pos, SEEK_SET);
             read_size = read(fd, buf, sizeof(buf));
             read_pos += read_size;
             
             /* Write whatever block we've got */
             lseek(fd, write_pos, SEEK_SET);
             write(fd, buf, read_size);
             write_pos += read_size;
          }
          while (read_size > 0);
       
          ftruncate(fd, stat_buf.st_size - current_totalsize);
    }
    
    close(fd);
    return 0;
}

/*
 * Function id3_write_tag (id3, fd)
 *
 *    Wrtite the ID3 tag to the indicated file descriptor.  Return 0
 *    upon success, or -1 if an error occured.
 *
 *    Warning: This function is called by id3_write_tag_filename and should
 *             not be called standalone, as it doesn't perform seeking
 *             and doesn't enlarge the file as necessary, and thus might
 *             OVERWRITE THE AUDIO DATA.
 *
 */
int id3_write_tag(struct id3_tag *id3, int fd)
{
	struct id3_frame *fr;
	GList *node;
	char buf[ID3_TAGHDR_SIZE];

	/*
	 * Write tag header.
	 */
	buf[0] = id3->id3_version;
	buf[1] = id3->id3_revision;
	buf[2] = id3->id3_flags;
	ID3_SET_SIZE28(id3->id3_size, buf[3], buf[4], buf[5], buf[6]);

	if (safe_write(fd, "ID3", 3) == -1)
		return -1;
	if (safe_write(fd, buf, ID3_TAGHDR_SIZE) == -1)
		return -1;

	/*
	 * TODO: Write extended header.
	 */
#if 0
	if (id3->id3_flags & ID3_THFLAG_EXT)
	{
		id3_exthdr_t exthdr;
	}
#endif

	for (node = id3->id3_frame; node != NULL; node = node->next)
	{
		char fhdr[ID3_FRAMEHDR_SIZE];

		fr = node->data;

		/*
		 * TODO: Support compressed headers, encoded
		 * headers, and grouping info.
		 */
		fhdr[0] = fr->fr_desc->fd_idstr[0];
		fhdr[1] = fr->fr_desc->fd_idstr[1];
		fhdr[2] = fr->fr_desc->fd_idstr[2];
		fhdr[3] = fr->fr_desc->fd_idstr[3];
		fhdr[4] = (fr->fr_raw_size >> 24) & 0xff;
		fhdr[5] = (fr->fr_raw_size >> 16) & 0xff;
		fhdr[6] = (fr->fr_raw_size >> 8) & 0xff;
		fhdr[7] = fr->fr_raw_size & 0xff;
		fhdr[8] = (fr->fr_flags >> 8) & 0xff;
		fhdr[9] = fr->fr_flags & 0xff;

		if (safe_write(fd, fhdr, sizeof(fhdr)) == -1)
			return -1;

		if (safe_write(fd, fr->fr_raw_data, fr->fr_raw_size) == -1)
			return -1;
	}
	return 0;
}

 
/*
 * Function id3_write_tag_file(id3, filename)
 *
 *   Write the ID3 tag to the file indicated by filename. Takes care of enlarging
 *   the file, if needed. Returns 0 upon success, or -1 if an error occured.
 *
 */
int id3_write_tag_filename(struct id3_tag *id3, const char* filename)
{
    struct id3_tag *current_id3; 
    int fd;
    int current_totalsize, new_totalsize;
    GList* node;
    
    fd = open(filename, O_RDWR);
    if (fd == -1)
       return -1;
      
    /*
     * Figure out how large the current tag is.
     */
    current_id3 = id3_open_fd(fd, 0);
    if (current_id3 != NULL)
    {
       /* We use MAX to make sure an erroneous tag doesn't confuse us */
       current_totalsize = MAX(current_id3->id3_totalsize, 0);
       id3_close(current_id3);
    }
    else
    {
       current_totalsize = 0;
    }
    
    /*
     * Figure out how large the new tag will be.
     */
    new_totalsize = 10;
    node = id3->id3_frame;
    while (node != NULL)
    {
	struct id3_frame* fr = node->data;
	
	{
		char* text = id3_get_text(fr);
		if (text != NULL)
		{
			int len = strlen(text);
			g_free(text);
			
			if (len == 0)
			{
				/* We skip over frames with empty text */
				node = node->next;
				id3_delete_frame(fr);
				continue;
			}
		}
	}
	
	{
		char* url  = id3_get_url(fr);
		if (url != NULL)
		{
			int len = strlen(url);
			g_free(url);
			
			if (len == 0)
			{
				/* We skip over frames with empty URLs */
				node = node->next;
				id3_delete_frame(fr);
				continue;
			}
		}
	}
	
	new_totalsize += fr->fr_raw_size + ID3_FRAMEHDR_SIZE;
	node = node->next;
    }
    new_totalsize += 0; /* no extended header, no footer, no padding */
    id3->id3_flags = 0;
    
    /*
     * Determine whether we need to rewrite the file to make place for the new tag.
     */
    if (new_totalsize > current_totalsize)
    {
       struct stat stat_buf;
       int grow_size;

       stat(filename, &stat_buf); 
       grow_size = new_totalsize - current_totalsize;
       ftruncate(fd, stat_buf.st_size + grow_size);
       
       /* truncate adds "sparse" zeros; we'll turn them into real ones,
          which occupy space on disk, to make sure we actually have
          the disk space before we start enlarging the file */
       {
          int remaining = grow_size;
          char buf[1024] = { 0 };
          lseek(fd, stat_buf.st_size, SEEK_SET);
          while (remaining > 0)
          {
             int ret = write(fd, buf, MIN(sizeof(buf), remaining));
             if (ret >= 0)
                remaining -= ret;
             else
             {
                id3_error(id3, "Unable to enlarge file for the new tag");
                ftruncate(fd, stat_buf.st_size);
                close(fd);
                return -1;
             }
          }
       }
       
       /* and now, re-position all the data grow_size bytes forward */
       {
          int area_to_move_size = stat_buf.st_size - current_totalsize;
          int read_pos, write_pos;
          size_t read_size, read_size_desired;
          char buf[4096];

          /* TODO: Add error handling to IO operations */
          
          /* We reposition the data by going backwards, each time reading
             a block and copying it forward. That way, we never write
             over a block which we intend to read later. */
          write_pos = lseek(fd, 0, SEEK_END);
          read_pos = write_pos - grow_size;

          /* While we still have bytes to move... */
          while(area_to_move_size > 0)
          {
             /* Get 1 block or the remaining bytes -- the smallest of the two */
             read_size_desired = MIN(area_to_move_size, sizeof(buf));
             read_pos -= read_size_desired;
             lseek(fd, read_pos, SEEK_SET);
             read_size = read(fd, buf, read_size_desired);
             
             /* Write whatever block we've got */
             write_pos -= read_size;
             lseek(fd, write_pos, SEEK_SET);
             write(fd, buf, read_size);
             
             area_to_move_size -= read_size;
          }
       }
    }
    else
    {
       new_totalsize = current_totalsize;
    }
    
    id3->id3_size = new_totalsize - 10;
    
    /* Zero-out the ID3v2 tag area */
    {
       char buf[1024] = {0};
       size_t remaining;
       
       lseek(fd, 0, SEEK_SET);
       for(remaining = new_totalsize; remaining > 0; remaining -= MIN(remaining,sizeof(buf)))
       {
          write(fd, buf, MIN(remaining,sizeof(buf)));
       }
    }

    /* Write the new tag */
    lseek(fd, 0, SEEK_SET);
    if (id3_write_tag(id3, fd) == -1)
    {
       close(fd);
       return -1;
    }
    
    close(fd);
    return 0;
}
