/*
 * Copyright (C) 1999-2001  Haavard Kvaalen <havardk@xmms.org>
 *
 * Licensed under GNU LGPL version 2.
 *
 */

#include <config.h>
#include <stdlib.h>
#include <glib.h>
#include <string.h>

#ifdef HAVE_ICONV
#include <iconv.h>
#endif

#include <errno.h>

#ifdef HAVE_CODESET
#include <langinfo.h>
#endif

#include "charset.h"

char* xmms_charset_get_current(void)
{
	char *charset = getenv("CHARSET");

#ifdef HAVE_CODESET
	if (!charset)
		charset = nl_langinfo(CODESET);
#endif
	/* Maybe we should default to ISO-8859-1 instead? */
	if (!charset)
		charset = "US-ASCII";

	return charset;
}

static size_t utf16_strlen(const char *string)
{
	size_t len = 0;

	if (!string)
		return 0;

	while (*(string + len) != 0 || *(string + len + 1) != 0)
		len += 2;

	return len;
}

#ifdef HAVE_ICONV
char* xmms_charset_convert(const char *string, size_t insize, char *from, char *to)
{
	size_t outleft, outsize;
	iconv_t cd;
	char *out, *outptr;
	ICONV_CONST char *input = (ICONV_CONST char *) string;

	if (!string)
		return NULL;

	if (!from)
		from = xmms_charset_get_current();
	if (!to)
		to = xmms_charset_get_current();

/* 	g_message("converting %s from %s to %s (%u)", string, from, to, insize); */
	if ((cd = iconv_open(to, from)) == (iconv_t)-1)
	{
		g_warning("convert_string(): Conversion not supported. "
			  "Charsets: %s -> %s", from, to);
		return g_strdup(string);
	}

	/* Due to a GLIBC bug, round outbuf_size up to a multiple of 4 */
	/* + 1 for nul in case len == 1 */
	outsize = ((insize + 3) & ~3) + 1;
	out = g_malloc(outsize);
	outleft = outsize - 1;
	outptr = out;

 retry:
	if (iconv(cd, &input, &insize, &outptr, &outleft) == -1)
	{
		int used;
		switch (errno)
		{
			case E2BIG:
				used = outptr - out;
				outsize = (outsize - 1) * 2 + 1;
				out = g_realloc(out, outsize);
				outptr = out + used;
				outleft = outsize - 1 - used;
				goto retry;
			case EINVAL:
				break;
			case EILSEQ:
				/* Invalid sequence, try to get the
                                   rest of the string */
				input++;
				insize--;
				goto retry;
			default:
				g_warning("convert_string(): Conversion failed. "
					  "Inputstring: %s; Error: %s",
					  string, strerror(errno));
				break;
		}
	}
	*outptr = '\0';

	iconv_close(cd);
	return out;
}

char* xmms_charset_from_utf8(const char *string)
{
	if (!string)
		return NULL;

	return xmms_charset_convert(string, strlen(string), "UTF-8", NULL);
}

char *xmms_charset_from_utf16(const unsigned char *string)
{
	if (!string)
		return NULL;

	return xmms_charset_convert(string, utf16_strlen(string), "UTF-16", NULL);
}

char *xmms_charset_from_utf16be(const unsigned char *string)
{
	if (!string)
		return NULL;

	return xmms_charset_convert(string, utf16_strlen(string), "UTF-16BE", NULL);
}

char* xmms_charset_from_latin1(const char *string)
{
	char *to = xmms_charset_get_current();

	if (!string)
		return NULL;

	if (!strcmp(to, "UTF-8"))
		return xmms_charset_convert(string, strlen(string), "ISO-8859-1", to);
	else
		return g_strdup(string);
}

#else

char* xmms_charset_convert(const char *string, size_t insize, char *from, char *to)
{
	if (!string)
		return NULL;

	if (!strcmp(from, "UTF-8") && !to)
		return xmms_charset_from_utf8(string);

	if (!strcmp(from, "UTF-16") && !to)
		return xmms_charset_from_utf16(string);

	if (!strcmp(from, "UTF-16BE") && !to)
		return xmms_charset_from_utf16be(string);

	return g_strdup(string);
}

char* xmms_charset_from_utf8(const char *string)
{
 	char *ascii, *c;
	const unsigned char *utf = string;
 	int len;

	if (!string)
		return NULL;

	len = strlen(string) + 1;
	ascii = g_malloc0(len);
	c = ascii;
	
	while (*utf != 0)
	{
		if (*utf < 0x80)
			*c++ = *utf++;
		else
		{
			char u = *utf++ << 1;
			*c++ = '?';
			
			/*
			 * Skip the entire utf8 character.
			 */
			while (u & 0x80 && *utf & 0x80 && !(*utf & 0x40))
			{
			       utf++;
			       u <<= 1;
			}
		}
	}
	return ascii;
}

static char* utf16_to_ascii(const unsigned char *utf16, int le)
{
	char *ascii;
	unsigned int i, len, c;

	if (!utf16)
		return NULL;

	len = utf16_strlen(utf16) / 2 + 1;
	
	ascii = g_malloc(len + 1);

	for (i = 0, c = 0; i < len; i++)
	{
		guint16 uc;
		int o = i << 1;

		if (le)
			uc = *(utf16 + o) | *(utf16 + o + 1) << 8;
		else
			uc = *(utf16 + o) << 8 | *(utf16 + o + 1);

		/* Skip BOM and surrogate pairs */
		if (uc == 0xfeff || (uc >= 0xd800 && uc <= 0xdfff))
			continue;
		
		if (uc < 0x80)
			ascii[c] = uc;
		else 
			ascii[c] = '?';
		c++;
	}
	
	ascii[c] = 0;
	return ascii;
}
	
char *xmms_charset_from_utf16(const unsigned char *string)
{
	int le = FALSE;
	guint16 bom;

	if (!string)
		return NULL;

	bom = *string << 8 | *(string + 1);

	if (bom == 0xfffe)
		le = TRUE;
	else if (bom != 0xfeff)
		return g_strdup("");

	return utf16_to_ascii(string, le);
}

char *xmms_charset_from_utf16be(const unsigned char *string)
{
	if (!string)
		return NULL;

	return utf16_to_ascii(string, FALSE);
}

char* xmms_charset_from_latin1(const char *string)
{
	if (!string)
		return NULL;

	return g_strdup(string);
}

#endif

char* xmms_charset_to_utf8(const char *string)
{
	if (!string)
		return NULL;

	return xmms_charset_convert(string, strlen(string), NULL, "UTF-8");
}
